from django.db.models import OneToOneField
from django.db.models.fields.related_descriptors import (
    ForwardOneToOneDescriptor,
    ReverseOneToOneDescriptor,
)


class NullableForwardOneToOneDescriptor(ForwardOneToOneDescriptor):

    def __get__(self, instance, cls=None):
        try:
            return super().__get__(instance, cls=cls)
        except self.RelatedObjectDoesNotExist:
            return None


class NullableReverseOneToOneDescriptor(ReverseOneToOneDescriptor):

    def __get__(self, instance, cls=None):
        try:
            return super().__get__(instance, cls=cls)
        except self.RelatedObjectDoesNotExist:
            return None


class NullableOneToOneField(OneToOneField):
    """ Modified version of Django's OneToOneField which sets the relation as nullable and doesn't
        raise DoesNotExist if the related object doesn't exist.
    """

    related_accessor_class = NullableReverseOneToOneDescriptor
    forward_related_accessor_class = NullableForwardOneToOneDescriptor

    def __init__(self, to, on_delete, to_field=None, **kwargs):
        kwargs["unique"] = True
        kwargs["null"] = True
        if "blank" not in kwargs:
            kwargs["blank"] = True
        super().__init__(to, on_delete, to_field=to_field, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "unique" in kwargs:
            del kwargs["unique"]
        if "null" in kwargs:
            del kwargs["null"]
        return name, path, args, kwargs



# Third Party
from django.core.exceptions import ValidationError
from django.db import models
import pydantic

# Fields of gold
from fields_of_gold import TypedJSONField
from .base import ExtraModelsTestCase


class SimpleType(pydantic.BaseModel):
    my_int: int
    my_str: str
    optional_int: int = 1
    optional_str: str = "String"


class SimpleTypedJSONModel(models.Model):

    class Meta:
        app_label = "fields_of_gold"

    typed_field = TypedJSONField(type=SimpleType)


class TypedJSONFieldTestCase(ExtraModelsTestCase):
    """ Tests for the TypedJSONField. """

    test_models = [SimpleTypedJSONModel]

    def test_basic_usage(self):
        instance = SimpleTypedJSONModel()
        instance.typed_field = SimpleType(my_int=1, my_str="cake")
        instance.typed_field.my_str = "not cake"
        instance.save()
        # To a full re-fetch so that we're testing a complete new instantiation from the DB
        instance = SimpleTypedJSONModel.objects.get(pk=instance.pk)
        self.assertIsInstance(instance.typed_field, SimpleType)
        self.assertEqual(instance.typed_field.my_int, 1)
        self.assertEqual(instance.typed_field.my_str, "not cake")

    def test_data_set_as_dict_is_converted(self):
        """ If the field's attribute value is set as a python dict, it should be converted to the
            Pydantic model type.
        """
        instance = SimpleTypedJSONModel()
        instance.typed_field = {"my_int": 1, "my_str": "hello"}
        instance.save()
        # As is the case with other Django fields (e.g. BooleanField), setting a value which is
        # valid but which is not the same type (e.g. setting 1 instead of True) will not get coerced
        # to the correct type simply by saving. It only gets coerced when fetching from the DB.
        instance.refresh_from_db()
        self.assertIsInstance(instance.typed_field, SimpleType)

    def test_invalid_data_raises_validation_error(self):
        """ Calling full_clean with invalid data should raise a Django ValidationError. """
        instance = SimpleTypedJSONModel()
        instance.typed_field = {"my_int": "Cannot be an int", "my_str": "whatever"}
        self.assertRaisesRegex(ValidationError, r"typed_field.+my_int", instance.full_clean)

    def test_non_nullable_field_raises_validation_error(self):
        instance = SimpleTypedJSONModel()
        self.assertRaisesRegex(
            ValidationError, r"typed_field.+cannot be null", instance.full_clean
        )
        instance.typed_field = SimpleType(my_int=1, my_str="cake")
        instance.full_clean()



# Standard library
from contextlib import contextmanager
import json
import logging

# Third Party
from django.core import checks
from django.core.exceptions import ValidationError
from django.db.models import JSONField
import pydantic


logger = logging.getLogger(__name__)


class PydanticJSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, pydantic.BaseModel):
            return o.model_dump(mode="json")
        return super().default(o)


class TypedJSONField(JSONField):
    """ A JSONField which can have an enforced schema using Pydantic models. """

    default_validators = [

    ]

    def __init__(self, *args, type=None, encoder=PydanticJSONEncoder, **kwargs):
        self.type = type
        super().__init__(*args, **{"encoder": encoder, **kwargs})

    def check(self, **kwargs):
        return [
            *super().check(),
            *self._check_type(),
        ]

    def _check_type(self):
        if not issubclass(self.type, pydantic.BaseModel):
            return [checks.Error(
                f"'type' argument must be a Pydantic model class, not {type(self.type)}.",
                obj=self,
                id="fields_of_gold.E001",
            )]
        return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["type"] = self.type
        return name, path, args, kwargs

    def to_python(self, value):
        """
        Convert the input value into the expected Python data type, raising
        django.core.exceptions.ValidationError if the data can't be converted.
        Return the converted value. Subclasses should override this.
        """
        # This relies on the top-level JSON object being a dict not a list
        if value is None or isinstance(value, self.type):
            return value

        try:
            return self.type(**value)
        except pydantic.ValidationError as error:
            raise ValidationError(str(error)) from error

    def from_db_value(self, value, expression, connection):
        value = super().from_db_value(value, expression, connection)
        try:
            return self.to_python(value)
        except ValidationError as error:
            logger.warning(
                "TypedJSONField got error loading value from DB for %s: %s. Raw data: %s",
                self.type,
                error,
                value,
            )
            # If there's a problem with the existing data in the DB then we want to just return the
            # raw JSON data as it is. If we don't do this, then bad existing data prevents the model
            # instance from being loaded at all.
            return value

    def validate(self, value, model_instance):
        """
        Validate value and raise ValidationError if necessary. Subclasses
        should override this to provide validation logic.
        """
        if isinstance(value, self.type):
            json_value = value.model_dump_json()
            super().validate(json_value, model_instance)
            with convert_validation_error():
                self.type.model_validate(value)
        else:
            super().validate(value, model_instance)
            if value is not None:
                with convert_validation_error():
                    self.type(**value)  # Will raise if data is invalid

    # def pre_save(self, model_instance, add):
    #     """ Enforce validation here??. """
    #     self.validate()
    #     pass

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, self.type):
            value = value.model_dump()
        return super().get_db_prep_value(value, connection, prepared)
        # if isinstance(value, self.type):
        #     return value.model_dump_json()
        # return value

    # We might need to override this
    # def value_to_string(self, obj):
    #     """
    #     Return a string value of this field from the passed obj.
    #     This is used by the serialization framework.
    #     """
    #     return str(self.value_from_object(obj))


@contextmanager
def convert_validation_error():
    try:
        yield
    except pydantic.ValidationError as error:
        raise ValidationError(str(error)) from error



# Third Party
from django.db import connection
from django.test import TestCase

class ExtraModelsTestCase(TestCase):
    """ Test case that sets up and destroys DB table(s) for additional models that are used only
        for the tests.
    """

    test_models = []

    @classmethod
    def setUpClass(cls):
        with connection.schema_editor() as schema_editor:
            for model in cls.test_models:
                schema_editor.create_model(model)
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        with connection.schema_editor() as schema_editor:
            for model in cls.test_models:
                schema_editor.delete_model(model)



import os
from setuptools import setup, find_packages


LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

VERSION = os.environ.get("GITHUB_REF", "v0.1.0").lstrip("refs/tags/").lstrip("v")

setup(
    name="fields-of-gold",
    version=VERSION,
    packages=find_packages(),
    description="A Django library for providing useful DB model fields.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author="Adam Alton",
    author_email="270255+adamalton@users.noreply.github.com",
    url="https://github.com/adamalton/fields-of-gold",
    keywords=["Django", "Pydantic", "JSON", "JSONField", "Schema"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    install_requires=[
        'django>=4.2,<6.0',
        'pydantic>2.0,<3.0',
    ],
    include_package_data=True,
    # extras_require={
    #     'test': [
    #     ]
    # },
)
